#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  4 10:52:17 2018

@author: tommy
"""
import numpy as np

def main():
    
    from KDEpy import FFTKDE, NaiveKDE
    from KDEpy.binning import linear_binning
    import matplotlib.pyplot as plt
    from scipy import stats
    
    np.random.seed(123)
    dist = stats.lognorm(1, 1)
    plt.figure(figsize=(14, 6))
    
    kernel = 'triweight'
    
    N = 10**3
    data = dist.rvs(int(N))
    plt.scatter(data, np.zeros_like(data), marker='|')
    x, y = NaiveKDE(bw='silverman', kernel=kernel).fit(data)(2**10)
    plt.plot(x, y, label='FFTKDE')
    plt.plot(x, dist.pdf(x), label='True')
    
    # -----------------------------------------------------------------------
    # Adaptive
    alpha = 1.9
    bw = 'silverman'
    kde = NaiveKDE(kernel='epa', bw=bw)
    kde.fit(data)(x)

    #y = NaiveKDE(bw=kde.bw*lambda_i).fit(x, weights=binned_data*lambda_i)(x)
    #plt.plot(x, y + np.ones_like(x)*0.00, label='Adaptive')
    
    # The FFTKDE grid may be wrong, but the true density cannot be
    # smaller than (1/N) K(0) at a given point
    min_kde = (1 / int(N)) * kde.kernel(0)
    kde_data = np.maximum(min_kde, kde(data))
    kde_data = kde(data)
    bw = kde.bw*((kde_data) / stats.mstats.gmean(kde_data))**-alpha
    print(np.min(kde(data)))
    print(stats.mstats.gmean(kde(data)))
    print(kde.bw, bw)
    #bw = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])/100
    plt.scatter(data, kde_data)
    y = NaiveKDE(bw=bw, kernel=kernel).fit(data, weights=None)(x)
    plt.plot(x, kde.bw*((kde(x) + 0) / stats.mstats.gmean(kde_data))**-alpha, label='bw')
    plt.plot(x, y + np.ones_like(x)*0.00, label='Adaptive')
    plt.ylim([0, 0.7])
    
    plt.legend()
    plt.show()
    
    print('-'*32)
    
    # -----------------------------------------------------------------------
    # Mirror at bounds
    plt.figure(figsize=(14, 6))
    
    # Beta distribution, where x=1 is a hard lower limit
    dist = stats.beta(a=1.05, b=3, loc=0, scale=1)
    
    # Plot the normal KDE and the true density
    data = dist.rvs(10**2)
    plt.figure(figsize=(14, 6))
    kde = FFTKDE(bw='silverman', kernel='triweight')
    x, y = kde.fit(data)(2**10)
    plt.figure(figsize=(14, 6))
    plt.plot(x, dist.pdf(x), label='True')
    plt.plot(x, y, label='FFTKDE')
    plt.scatter(data, np.zeros_like(data), marker='|')
    print(np.min(data), np.max(data))
    
    data_transformed = np.log(data)
    plt.scatter(data_transformed, np.zeros_like(data_transformed), marker='|')
    kde = FFTKDE(bw='silverman', kernel='triweight')
    x, y = kde.fit(data_transformed)(2**10)
    plt.plot(x, y, label='FFTKDE - transformed')
    
    print(x)
    print(y)
    plt.plot(np.exp(x), 2*np.exp(y)*(1 + y) - 2)
    
    plt.ylim([0, 3])
    plt.xlim([-1, 4])
    
    plt.legend()
    plt.show()
    
    
    # -------------------------------------------------------------------------
    # Data on a circle
    # Beta distribution, where x=1 is a hard lower limit
    np.random.seed(123)
    
    dist1 = stats.norm(loc=0, scale=1)
    dist2 = stats.norm(loc=20, scale=1)
    dist3 = stats.norm(loc=40, scale=1)
    data = np.hstack([dist1.rvs(10**3), dist2.rvs(10**3), dist3.rvs(10**3)])


    plt.figure(figsize=(14, 6))
    x, y = FFTKDE(bw='silverman').fit(data)()
    plt.plot(x, (dist1.pdf(x) + dist2.pdf(x)+ dist3.pdf(x)) / 3, label='True distribution')
    plt.plot(x, y, label="FFTKDE with Silverman's rule")

    y = FFTKDE(bw='ISJ').fit(data)(x)
    plt.plot(x, y, label="FFTKDE with Improved Sheather Jones (ISJ)")

    
    plt.legend()
    plt.show()
    
    
    
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  4 10:52:17 2018

@author: tommy
"""
import numpy as np
import numbers


def cartesian(arrays):
    """
    Generate a cartesian product of input arrays.
    Adapted from: 
        https://github.com/scikit-learn/scikit-learn/blob/
        master/sklearn/utils/extmath.py#L489
        
    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.
        
    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.
        
    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])
    """
    arrays = [np.asarray(x) for x in arrays]
    shape = (len(x) for x in arrays)
    dtype = arrays[0].dtype

    ix = np.indices(shape)
    ix = ix.reshape(len(arrays), -1).T

    out = np.empty_like(ix, dtype=dtype)

    for n, arr in enumerate(arrays):
        out[:, n] = arrays[n][ix[:, n]]

    return out


def autogrid(data, boundary_abs=3, num_points=None, boundary_rel=0.05):
    """
    Automatically select a grid if the user did not supply one.
    Input is (obs, dims), and so is ouput.
    
    number of grid : should be a power of two
    percentile : is how far out we go out
    
    Parameters
    ----------
    data : array-like
        Data with shape (obs, dims).
    boundary_abs: float
        How far out from boundary observations the grid goes in each dimension.
    num_points: int
        The number of points in the resulting grid (after cartesian product).
        Should be a number such that k**dims = `num_points`.
    boundary_rel: float
        How far out to go, relatively to max - min.
        
    Returns
    -------
    grid : array-like
        A grid of shape (obs, dims).
        
    Examples
    --------
    >>> autogrid(np.array([[0, 0]]), boundary_abs=1, num_points=3)
    array([[-1., -1.],
           [-1.,  0.],
           [-1.,  1.],
           [ 0., -1.],
           [ 0.,  0.],
           [ 0.,  1.],
           [ 1., -1.],
           [ 1.,  0.],
           [ 1.,  1.]])
    >>> autogrid(np.array([[0, 0]]), boundary_abs=0.5, num_points=(2, 3))
    array([[-0.5, -0.5],
           [-0.5,  0. ],
           [-0.5,  0.5],
           [ 0.5, -0.5],
           [ 0.5,  0. ],
           [ 0.5,  0.5]])
    """
    obs, dims = data.shape
    minimums, maximums = data.min(axis=0), data.max(axis=0)
    ranges = maximums - minimums
    
    if num_points is None:
        num_points = [int(np.power(1024, 1 / dims))] * dims
    elif isinstance(num_points, (numbers.Number,)):
        num_points = [num_points] * dims
    elif isinstance(num_points, (list, tuple)):
        pass
    else:
        msg = '`num_points` must be None, a number, or list/tuple for dims'
        raise TypeError(msg)
        
    if not len(num_points) == dims:
        raise ValueError('Number of points must be sequence matching dims.')
    
    list_of_grids = []

    generator = enumerate(zip(minimums, maximums, ranges, num_points))
    for i, (minimum, maximum, rang, points) in generator:
        assert points >= 2
        outside_borders = max(boundary_rel * rang, boundary_abs)
        list_of_grids.append(np.linspace(minimum - outside_borders,
                                         maximum + outside_borders,
                                         num=points))

    return cartesian(list_of_grids)
    

if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v', '--capture=sys'])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  4 10:52:17 2018

@author: tommy
"""
import numbers
import itertools
import numpy as np
from KDEpy.BaseKDE import BaseKDE


class NaiveKDE(BaseKDE):
    """
    This class implements a naive computation of a kernel density estimate. The
    advantages are that choices of bandwidth, norms, weights and grids are
    straightforward -- the user can do almost anything. The disadvantage is 
    that computations are slow on more than a couple of thousand data points.

    Parameters
    ----------
    kernel : str
        The kernel function. See cls._available_kernels.keys() for choices.
    bw : float, str or array-like
        Bandwidth or bandwidth selection method. If a float is passed, it
        is the standard deviation of the kernel. If a string it passed, it
        is the bandwidth selection method, see cls._bw_methods.keys() for
        choices. If an array-like it passed, it is the bandwidth of each
        point.
    norm : float
        The p-norm used to compute the distances in higher dimensions.
        
    Examples
    --------
    >>> data = np.random.randn(2**10)
    >>> # (1) Automatic bw selection using Improved Sheather Jones
    >>> x, y = NaiveKDE(bw='ISJ').fit(data).evaluate()
    >>> # (2) Explicit choice of kernel and bw (standard deviation of kernel)
    >>> x, y = NaiveKDE(kernel='triweight', bw=0.5).fit(data).evaluate()
    >>> weights = data + 10
    >>> # (3) Using a grid and weights for the data
    >>> y = NaiveKDE(kernel='epa', bw=0.5).fit(data, weights).evaluate(x)
    
    References
    ----------
    - Silverman, B. W. Density Estimation for Statistics and Data Analysis. 
      Boca Raton: Chapman and Hall, 1986.
    - Wand, M. P., and M. C. Jones. Kernel Smoothing. 
      London ; New York: Chapman and Hall/CRC, 1995.
    - Scipy implementation, at ``scipy.stats.gaussian_kde``.
    """
    
    def __init__(self, kernel='gaussian', bw=1, norm=2):
        super().__init__(kernel, bw)
        self.norm = norm
    
    def fit(self, data, weights=None):
        """
        Fit the KDE to the data. This validates the data and stores it. 
        Computations are performed when the KDE is evaluated on a grid.
    
        Parameters
        ----------
        data: array-like
            The data points. High dimensional data must have shape (obs, dims).
        weights: array-like
            One weight per data point. Must have shape (obs,). If None is
            passed, uniform weights are used.
            
        Returns
        -------
        self
            Returns the instance.
            
        Examples
        --------
        >>> data = [1, 3, 4, 7]
        >>> weights = [3, 4, 2, 1]
        >>> kde = NaiveKDE().fit(data, weights=None)
        >>> kde = NaiveKDE().fit(data, weights=weights)
        >>> x, y = kde()
        """
        # Sets self.data
        super().fit(data, weights)
        return self
    
    def evaluate(self, grid_points=None):
        """
        Evaluate on grid points.
        
        Parameters
        ----------
        grid_points: array-like, int, tuple or None
            A grid (mesh) to evaluate on. High dimensional grids must have 
            shape (obs, dims). If an integer is passed, it's the number of grid
            points on an equidistant grid. If a tuple is passed, it's the
            number of grid points in each dimension. If None, a grid will be 
            automatically created.
            
        Returns
        -------
        y: array-like
            If a grid is supplied, `y` is returned. If no grid is supplied,
            a tuple (`x`, `y`) is returned.
            
        Examples
        --------
        >>> kde = NaiveKDE().fit([1, 3, 4, 7])
        >>> # Two ways to evaluate, either with a grid or without
        >>> x, y = kde.evaluate()
        >>> x, y = kde.evaluate(256)
        >>> y = kde.evaluate(x)
        """
        # This method sets self.grid points and verifies it
        # NaiveKDE does not convert the bw to a scalar, since a vector is
        # allowed too.
        super().evaluate(grid_points, bw_to_scalar=False)

        # Create zeros on the grid points
        evaluated = np.zeros(self.grid_points.shape[0])
        
        # For every data point, compute the kernel and add to the grid
        bw = self.bw
        if isinstance(bw, numbers.Number):
            bw = np.asfarray(np.ones(self.data.shape[0]) * bw)
        elif callable(bw):
            bw = np.asfarray(np.ones(self.data.shape[0]) * bw(self.data))

        # TODO: Implementation w.r.t grid points for faster evaluation  
        weights = (itertools.repeat(1 / self.data.shape[0]) if self.weights 
                   is None else self.weights)
            
        for weight, data_point, bw in zip(weights, self.data, bw):
            x = self.grid_points - data_point
            evaluated += weight * self.kernel(x, bw=bw, norm=self.norm)

        return self._evalate_return_logic(evaluated, self.grid_points)


if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v'])
    

if __name__ == "__main__":
    
    import matplotlib.pyplot as plt
    
    # Basic example of the naive KDE
    # -----------------------------------------
    data = [3, 3.5, 4, 6, 8]
    kernel = 'gaussian'
    bw = 1
    
    plt.figure(figsize=(10, 4))
    plt.title('Basic example of the naive KDE')
    
    plt.subplot(1, 2, 1)
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data)
    x = np.linspace(0, 10, num=1024)
    for d in data:
        k = NaiveKDE(kernel=kernel, bw=bw).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
      
    plt.subplot(1, 2, 2)
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data)
    x = np.linspace(0, 10, num=1024)
    for d in data:
        k = NaiveKDE(kernel=kernel, bw=bw).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    plt.show()
    
    # Naive KDE with weights
    # -----------------------------------------
    data = [3, 3.5, 4, 6, 8]
    weights = np.array([1, 1, 1, 1, 5])
    weights = weights / np.sum(weights)
    kernel = 'gaussian'
    bw = 1
    
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights=weights)
    
    x = np.linspace(0, 10, num=1024)
    for d, w in zip(data, weights):
        k = (NaiveKDE(kernel=kernel, bw=bw).fit([d], weights=[w]).evaluate(x) *
             w)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)
    plt.title('Naive KDE with weights')
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    plt.show()
    
    # Naive KDE with variable h
    # -----------------------------------------
    data = [2, 3, 4, 5, 6, 7]
    bws = [1, 2, 3, 4, 5, 6]
    bws = [1 / k for k in bws]
    kernel = 'gaussian'
    
    kde = NaiveKDE(kernel=kernel, bw=bws)
    kde.fit(data)
    
    x = np.linspace(0, 10, num=1024)
    for d, bw in zip(data, bws):
        k = NaiveKDE(kernel=kernel, bw=bw).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)  
    plt.title('Naive KDE with variable h')
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    plt.show()
    
    # Naive KDE with silverman
    # -----------------------------------------
    data = [2, 3, 4, 5, 6, 7]
    bws = [1, 2, 3, 4, 5, 6]
    bws = [1 / k for k in bws]
    kernel = 'gaussian'
    
    kde = NaiveKDE(kernel=kernel, bw='silverman')
    kde.fit(data)
    
    x = np.linspace(0, 10, num=1024)
    for d, bw in zip(data, bws):
        k = (NaiveKDE(kernel=kernel, bw='silverman').fit([d]).evaluate(x) / 
             len(data))
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)  
    plt.title('Naive KDE with silverman')
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    plt.show()



# -*- coding: utf-8 -*-
"""
Fast cython functions for linear binning.

Notes
-----
(1) Instead of computing the integral and fractional part of a number as
    integral = int(data_point)
    fractional = data_point % 1
using the following code is x4 times faster:
    integral = int(data_point)
    fractional = data_point - integral
    
(2) It is extremely important to type EVERYTHING with cdef or in the function
signature. If even one variable is not properly typed, much of the speed gain
is gone.
"""

cimport cython
from libc.math cimport floor

# boundscheck(False) -> Cython is free to assume that indexing will not cause 
# any IndexErrors to be raised.

# wraparound(False) ->  If set to False, Cython is allowed to neither check 
# for nor correctly handle negative indices

# cdivision(True) -> If set to False, Cython will adjust the remainder and 
# quotient operators C types to match those of Python ints (which differ 
# when the operands have opposite signs) and raise a ZeroDivisionError 
# when the right operand is 0
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_1D_weighted(double[:] transformed_data, double[:] weights, double[:] result):
    """
    1D fast binning with weights. Faster than N-dimensional cython function
    because it unrolls the loops.
    
    Parameters
    ----------
    transformed_data : the transformed (scaled w.r.t grid) data
    weights : the weights
    result : array to put the results of the computation in
    """
    cdef int obs, integral
    cdef double data_point, weight, fractional, frac_times_weight
    
    obs = transformed_data.shape[0]

    for i in range(obs):
        data_point = transformed_data[i]
        weight = weights[i]
        # Using <int> as shown below is faster than int( . ) 
        # integral = int(data_point)
        integral = <int> floor(data_point)
        fractional = data_point - integral
        frac_times_weight = fractional * weight  # Compute product once
        result[integral + 1] += frac_times_weight
        result[integral] += weight - frac_times_weight

    return result


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_1D(double[:] transformed_data, double[:] result):
    """
    1D fast binning weights. Faster than N-dimensional cython function
    because it unrolls the loops.
    
    Parameters
    ----------
    transformed_data : the transformed (scaled w.r.t grid) data
    result : array to put the results of the computation in
    """
    cdef int obs, integral
    cdef double data_point, weight, fractional
    obs = transformed_data.shape[0]
    for i in range(obs):
        data_point = transformed_data[i]
        integral = <int> floor(data_point)
        fractional = data_point - integral
        result[integral] += 1 - fractional
        result[integral + 1] += fractional

    return result


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_2D_weighted(double[:, :] data, double[:] weights, 
                             double[:] result, long[:] grid_num, int obs_tot):
    """
    2D fast binning with weights. Faster than N-dimensional cython function
    because it unrolls the loops.
    
    Parameters
    ----------
    data : the transformed (scaled w.r.t grid) data
    weights : the weights
    result : array to put the results of the computation in
    grid_num : number of grid points in each dimension
    obs_tot : total number of observations (grid points)
    """
    cdef int obs, index, i, x_integral, y_integral, grid_num1
    cdef double x, y, weight, x_fractional, y_fractional, value, xy, y_xy, x_xy
    
    obs = data.shape[0]
    for i in range(obs):
        
        x = data[i, 0]
        y = data[i, 1]
        weight = weights[i]
        
        x_integral = <int> floor(x)
        x_fractional = x - x_integral
        y_integral = <int> floor(y)
        y_fractional = y - y_integral

        # Computations with few flops
        xy = x_fractional * y_fractional
        y_xy = y_fractional - xy
        x_xy = x_fractional - xy
        grid_num1 = grid_num[1]
        # Bottom left
        index = y_integral + x_integral * grid_num1
        result[index % obs_tot] += (xy - x_fractional - y_fractional + 1) * weight
        
        # Bottom right
        index = y_integral + (x_integral + 1) * grid_num1
        result[index % obs_tot] += x_xy * weight
        
        # Top left
        index = (y_integral + 1) + x_integral * grid_num1
        result[index % obs_tot] += y_xy * weight
        
        # Top right
        index = (y_integral + 1) + (x_integral + 1) * grid_num1
        result[index % obs_tot] += xy * weight
        
    return result

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_2D(double[:, :] data, double[:] result, long[:] grid_num, 
                    int obs_tot):
    """
    2D fast binning. Faster than N-dimensional cython function because it 
    unrolls the loops. See `iterate_data_2D_weighted` for commented code.
    
    Parameters
    ----------
    data : the transformed (scaled w.r.t grid) data
    weights : the weights
    result : array to put the results of the computation in
    grid_num : number of grid points in each dimension
    obs_tot : total number of observations (grid points)
    """
    cdef int obs, index, i, x_integral, y_integral, grid_num1
    cdef double x, y, x_fractional, y_fractional, value, xy, y_xy, x_xy
    
    obs = data.shape[0]
    for i in range(obs):
        x = data[i, 0]
        y = data[i, 1]
        x_integral = <int> floor(x)
        x_fractional = x - x_integral
        y_integral = <int> floor(y)
        y_fractional = y - y_integral
        xy = x_fractional * y_fractional
        y_xy = y_fractional - xy
        x_xy = x_fractional - xy
        grid_num1 = grid_num[1] 
        index = y_integral + x_integral * grid_num1
        result[index % obs_tot] += (xy - x_fractional - y_fractional + 1)
        index = y_integral + (x_integral + 1) * grid_num1
        result[index % obs_tot] += x_xy
        index = (y_integral + 1) + x_integral * grid_num1
        result[index % obs_tot] += y_xy
        index = (y_integral + 1) + (x_integral + 1) * grid_num1
        result[index % obs_tot] += xy

    return result


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_ND(double[:, :] data, double[:] result, long[:] grid_num, 
                    int obs_tot, long[:, :] binary_flgs):
    """
    Iterate over N-dimensional data and bin it.
    
    The idea behind this N-dimensional generalization is to pre-compute binary 
    numbers up to 2**dims. E.g. (0,0,0), (0,0,1), (0,1,0), (0,1,1), .. for 3 
    dimensions. Each tuple represent a corner in N-space. Let t_k be the tuple 
    binary value at index k, then the index computation and the value at the
    grid point may be expressed as:
        
        index = sum_i^n (x_i prod_j=i+1^n g_j)
                where x_i := (int(x_i) + 0^t_k) 
        
        value = prod_i^n (1 - (x_i % 1))^t_k * (x_i % 1)^(t_k - 1)
    
    Parameters
    ----------
    data : the transformed (scaled w.r.t grid) data
    result : array to put the results of the computation in
    grid_num : number of grid points in each dimension
    obs_tot : total number of observations (grid points)
    binary_flgs : array of shape (dims, 2**dims), counting in binary
                 this is used to to go every corner point efficiently
    """
    cdef int obs, result_index, i, dims, corners, j, flg, corner, integer_xi
    cdef double corner_value, fraction
    cdef double[:] x_i
    
    # Get the observations and dimensions of the data
    obs, dims = data.shape[0], data.shape[1]
    
    # For every dimension, there are two directions to find corners in
    corners = 2**dims

    # Loop through every data point
    for i in range(obs):
        
        # Retrieve the data point to consider
        x_i = data[i, :]
        
        # The data point will be 'assigned' to the 2**dims corners of the grid
        # that are closed to it. To do this, we loop through every corner
        for corner in range(corners):
            
            # For this corner, we must find the index (1) of the `result` array
            # to input the computed result, as well as the actual result (2).
            
            # (1) To compute the index of this corner, and the value, we must
            # again loop through x_1, x_2, ..., d_x. If g_i is the number of
            # grid points in the i'th dimension, the index is found by
            # x_1 * (g_2 * g_3 * g_4) + x_2 * (g_3 * g_4) + x_3 * (g_4) + x_4
            # = g_4( g_3(g_2(x_1) + x_2) + x_3) + x_4
            # Since we use flags to indicate x_1 or x_1 + 1, the following
            # code does the job:
            result_index = int(x_i[0])
            result_index += 0**binary_flgs[corner, 0]
            for j in range(1, dims):
                result_index *= grid_num[j]
                integer_xi = <int> floor(x_i[j])
                result_index += (integer_xi + 0**binary_flgs[corner, j])
                
            # (2) The value is found by
            # PROD_{i=0} (1 - frac(x[i))**flg * frac(x[i]) ** (1 - flg)
            corner_value = 1.0
            for j in range(dims):
                flg = binary_flgs[corner, j]
                # Compute this part of the product, using binary flags to
                # indicate whether the factor should be x_i or (1 - x_i)
                fraction = x_i[j] % 1
                corner_value *= (1 - fraction)**flg * fraction**(1 - flg)
            
            # Finished computing index and result, add to the grid corner point
            result[result_index % obs_tot] += corner_value
    
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def iterate_data_ND_weighted(double[:, :] data, double[:] weights, double[:] result, 
                    long[:] grid_num, int obs_tot, long[:, :] binary_flgs):
    """
    See `iterate_data_ND` for documentation.
    """
    cdef int obs, result_index, i, dims, corners, j, flg, corner, integer_xi
    cdef double corner_value, fraction, weight
    cdef double[:] x_i
    
    obs, dims = data.shape[0], data.shape[1]
    corners = 2**dims
    
    for i in range(obs):
        x_i = data[i, :]
        weight = weights[i]

        for corner in range(corners):
            
            result_index = int(x_i[0])
            result_index += 0**binary_flgs[corner, 0]
            for j in range(1, dims):
                result_index *= grid_num[j]
                integer_xi = <int> floor(x_i[j])
                result_index += (integer_xi + 0**binary_flgs[corner, j])

            corner_value = 1.0
            for j in range(dims):
                flg = binary_flgs[corner, j]
                fraction = x_i[j] % 1
                corner_value *= (1 - fraction)**flg * fraction**(1 - flg)
                
            result[result_index % obs_tot] += corner_value  * weight
    
    return result


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions for bandwidth selection.
"""
import numpy as np
import scipy
from KDEpy.binning import linear_binning
from KDEpy.utils import autogrid
from scipy import fftpack
from scipy.optimize import brentq

try:
    FLOAT = scipy.float128
except AttributeError:
    FLOAT = np.float64


def _fixed_point(t, N, I_sq, a2):
    r"""
    Compute the fixed point as described in the paper by Botev et al.
    
    .. math:
        
        t = \xi \gamma^{5}(t)
    
    Parameters
    ----------
    t : float
        Initial guess.
    N : int
        Number of data points.
    I_sq : array-like
        The numbers [1, 2, 9, 16, ...]
    a2 : array-like
        The DCT of the original data, divided by 2 and squared.
    
    Examples
    --------
    >>> # From the matlab code
    >>> ans = _fixed_point(0.01,50,np.arange(1, 51),np.arange(1, 51))
    >>> assert np.allclose(ans, 0.009947962622371)
    >>> # another
    >>> ans = _fixed_point(0.07,25,np.arange(1, 11),np.arange(1, 11))
    >>> assert np.allclose(ans, 0.069100181315957)
    
    References
    ----------
     - Implementation by Daniel B. Smith, PhD, found at
       https://github.com/Daniel-B-Smith/KDE-for-SciPy/blob/master/kde.py
    """
    
    # This is important, as the powers might overflow if not done
    I_sq = np.asfarray(I_sq, dtype=FLOAT)
    a2 = np.asfarray(a2, dtype=FLOAT)
    
    # ell = 7 corresponds to the 5 steps recommended in the paper
    ell = 7
    
    # Fast evaluation of |f^l|^2 using the DCT, see Plancherel theorem
    f = 2 * np.pi**(2 * ell) * np.sum(np.power(I_sq, ell) * a2 * 
                                      np.exp(-I_sq * np.pi**2 * t))

    # Norm of a function, should never be negative
    if f <= 0:
        return -1
    for s in reversed(range(2, ell)):
        # This could also be formulated using the double factorial n!!,
        # but this is faster so and requires an import less
        
        # Step one: estimate t_s from |f^(s+1)|^2
        odd_numbers_prod = np.product(np.arange(1, 2 * s + 1, 2, dtype=FLOAT))
        K0 = odd_numbers_prod / np.sqrt(2 * np.pi)
        const = (1 + (1 / 2) ** (s + 1 / 2)) / 3
        time = np.power((2 * const * K0 / (N * f)), 
                        (2. / (3. + 2. * s)))
        
        # Step two: estimate |f^s| from t_s
        f = 2 * np.pi**(2 * s) * np.sum(np.power(I_sq, s) * a2 * 
                                        np.exp(-I_sq * np.pi**2 * time))
        
    # This is the minimizer of the AMISE
    t_opt = np.power(2 * N * np.sqrt(np.pi) * f, -2. / 5)
    
    # Return the difference between the original t and the optimal value
    return t - t_opt


def _root(function, N, args):
    """
    Root finding algorithm. Based on MATLAB implementation by Botev et al.
    
    >>> # From the matlab code
    >>> ints = np.arange(1, 51)
    >>> ans = _root(_fixed_point, N=50, args=(50, ints, ints))
    >>> np.allclose(ans, 5.203713947289470e-05)
    True
    """
    # From the implementation by Botev, the original paper author
    # Rule of thumb of obtaining a feasible solution
    N = max(min(1050, N), 50)
    tol = 10e-12 + 0.01 * (N - 50) / 1000
    # While a solution is not found, increase the tolerance and try again
    found = 0
    while found == 0:
        try:
            # Other viable solvers include: [brentq, brenth, ridder, bisect]
            x, res = brentq(function, 0, tol, args=args, 
                            full_output=True, disp=False)
            found = 1 if res.converged else 0
        except ValueError:
            x = 0
            tol *= 2.
            found = 0
        if x <= 0:
            found = 0

        # If the tolerance grows too large, minimize the function
        if tol >= 1:
            raise ValueError('Root finding did not converge. Need more data.')
            
    if not x > 0:
        raise ValueError('Root finding failed to find positive solution.')
    return x


def improved_sheather_jones(data):
    """
    The Improved Sheater Jones (ISJ) algorithm from the paper by Botev et al.
    This algorithm computes the optimal bandwidth for a gaussian kernel,
    and works very well for bimodal data (unlike other rules). The
    disadvantage of this algorithm is longer computation time, and the fact
    that this implementation does not always converge if very few data
    points are supplied.

    Understanding this algorithm is difficult, see:
    https://books.google.no/books?id=Trj9HQ7G8TUC&pg=PA328&lpg=PA328&dq=
    sheather+jones+why+use+dct&source=bl&ots=1ETdKd_6EF&sig=jZk4R515GB1xsn-
    VZVnjr-JfjSI&hl=en&sa=X&ved=2ahUKEwi1_czNncTcAhVGhqYKHaPiBtcQ6AEwA3oEC
    AcQAQ#v=onepage&q=sheather%20jones%20why%20use%20dct&f=false
    """
    obs, dims = data.shape
    assert dims == 1
    
    n = 2**10
    # Setting `percentile` higher decreases the chance of overflow
    xmesh = autogrid(data, boundary_abs=6, num_points=n, boundary_rel=0.5)
    data = data.ravel()
    xmesh = xmesh.ravel()
    
    # Create an equidistant grid
    R = np.max(data) - np.min(data)
    # dx = R / (n - 1)
    data = data.ravel()
    N = len(np.unique(data))

    # Use linear binning to bin the data on an equidistant grid, this is a
    # prerequisite for using the FFT (evenly spaced samples)
    initial_data = linear_binning(data.reshape(-1, 1), xmesh)
    assert np.allclose(initial_data.sum(), 1)
    
    # Compute the type 2 Discrete Cosine Transform (DCT) of the data
    a = fftpack.dct(initial_data)
    
    # Compute the bandwidth
    I_sq = np.power(np.arange(1, n, dtype=FLOAT), 2)
    a2 = a[1:]**2 / 4

    # Solve for the optimal (in the AMISE sense) t
    t_star = _root(_fixed_point, N, args=(N, I_sq, a2))

    # The remainder of the algorithm computes the actual density
    # estimate, but this function is only used to compute the 
    # bandwidth, since the bandwidth may be used for other kernels
    # apart from the Gaussian kernel

    # Smooth the initial data using the computed optimal t  
    # Multiplication in frequency domain is convolution   
    # integers = np.arange(n, dtype=np.float)
    # a_t = a * np.exp(-integers**2 * np.pi ** 2 * t_star / 2)

    # Diving by 2 done because of the implementation of fftpack.idct
    # density = fftpack.idct(a_t) / (2 * R)
    
    # Due to overflow, some values might be smaller than zero, correct it
    # density[density < 0] = 0.
    bandwidth = np.sqrt(t_star) * R
    return bandwidth


def scotts_rule(data):
    """
    Scotts rule.
    
    Scott (1992, page 152)
    Scott, D.W. (1992) Multivariate Density Estimation. Theory, Practice and 
    Visualization. New York: Wiley. 
    
    Examples
    --------
    >>> data = np.arange(9).reshape(-1, 1)
    >>> ans = scotts_rule(data)
    >>> assert np.allclose(ans, 1.76474568962182)
    """
    if not len(data.shape) == 2:
        raise ValueError('Data must be of shape (obs, dims).')
    
    obs, dims = data.shape
    assert dims == 1
    sigma = np.std(data, ddof=1)
    # scipy.norm.ppf(.75) - scipy.norm.ppf(.25) -> 1.3489795003921634
    IQR = ((np.percentile(data, q=75) - np.percentile(data, q=25)) / 
           1.3489795003921634)

    sigma = min(sigma, IQR)
    return sigma * np.power(obs, -1. / (dims + 4))


def silvermans_rule(data):
    """
    Returns optimal smoothing (standard deviation) if the data is close to 
    normal.
    
    TODO: Extend to multidimensional:
        https://docs.scipy.org/doc/scipy-0.13.0/reference/generated/scipy.
        stats.gaussian_kde.html#r216
        
    Examples
    --------
    >>> data = np.arange(9).reshape(-1, 1)
    >>> ans = silvermans_rule(data)
    >>> assert np.allclose(ans, 1.8692607078355594)
    """
    if not len(data.shape) == 2:
        raise ValueError('Data must be of shape (obs, dims).')
    obs, dims = data.shape
    assert dims == 1

    if obs == 1:
        return 1
    if obs < 1:
        raise ValueError('Data must be of length > 0.')
        
    sigma = np.std(data, ddof=1)
    # scipy.norm.ppf(.75) - scipy.norm.ppf(.25) -> 1.3489795003921634
    IQR = ((np.percentile(data, q=75) - np.percentile(data, q=25)) / 
           1.3489795003921634)

    sigma = min(sigma, IQR)
    return sigma * (obs * 3 / 4.) ** (-1 / 5)

    
_bw_methods = {'silverman': silvermans_rule,
               'scott': scotts_rule,
               'ISJ': improved_sheather_jones}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from KDEpy.NaiveKDE import NaiveKDE
from KDEpy.TreeKDE import TreeKDE
from KDEpy.FFTKDE import FFTKDE

__version__ = '0.5.6'
__author__ = 'tommyod'

TreeKDE = TreeKDE
NaiveKDE = NaiveKDE
FFTKDE = FFTKDE



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for functions related to linear binning. These functions employ
linear binning to weighted data. This is typically a preprocessing step
before convolving with a kernel in the FFTKDE, but may also be used to
reduce the effective number of data points in any algorithm.

The idea behind linear binning is the following: (1) go through every
data point and (2) assign weight to the 2^dims nearest grid points.
In `dims` dimensions, there are 2 points on the grid to consider in
each direction, so a total of 2^dims grid points to assign weights to
for every data point. The weights are determined by the proportion of
the volume of this hypercube that is enclosed by the data point.

A ------------------------------ B
|                    |           |
|                    |           |
|                    X-----------|
|                                |
|                                |
|                                |
C ------------------------------ C

References
----------
- Fan, Jianqing, and James S. Marron. 
  “Fast Implementations of Nonparametric Curve Estimators.” 
  Journal of Computational and Graphical Statistics 3, no. 1 (March 1, 1994). 
  https://doi.org/10.1080/10618600.1994.10474629.
"""
import numpy as np
import itertools
import functools
import operator
from KDEpy.utils import cartesian

# Modulenotfounderror does not exist in pyton 3.5, fails on linter
import sys
if sys.version_info.minor == 6:
    ModuleNotFoundError = None

try:
    import cutils
    _use_Cython = True
except ModuleNotFoundError:
    _use_Cython = False


def linbin_cython(data, grid_points, weights=None):
    """
    1D Linear binning using Cython. Assigns weights to grid points from data.
    
    Runs in approx 10 ms on 1 million data points.
    
    Parameters
    ----------
    data : array-like
        Should be of shape (obs,).
    grid_points : array-like
        Should be of shape (points,).
    weights : array-like
        Should be of shape (obs,).
        
    Examples
    --------
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_cython(data, np.arange(6), weights=None)
    >>> np.allclose(ans, np.array([0, 0, 0.375, 0.375, 0.25, 0]))
    True
    >>> ans = linbin_cython(data, np.arange(6), weights=np.arange(1, 5))
    >>> np.allclose(ans, np.array([0, 0, 0.2, 0.4, 0.4, 0]))
    True
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_cython(data, np.arange(1, 7), weights=None)
    >>> np.allclose(ans, np.array([0, 0.375, 0.375, 0.25, 0, 0]))
    True
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=np.float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=np.float)

    # Verify that the grid is equidistant
    diffs = np.diff(grid_points)
    assert np.allclose(np.ones_like(diffs) * diffs[0], diffs)

    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=np.float)
        weights = weights / np.sum(weights)

    if (weights is not None) and (len(data) != len(weights)):
        raise ValueError('Length of data must match length of weights.')

    # Transform the data
    min_grid = np.min(grid_points)
    max_grid = np.max(grid_points)
    num_intervals = len(grid_points) - 1  # Number of intervals
    dx = (max_grid - min_grid) / num_intervals
    transformed_data = (data - min_grid) / dx

    result = np.asfarray(np.zeros(num_intervals + 2))

    if weights is None:
        result = cutils.iterate_data_1D(transformed_data, result)
        return np.asfarray(result[:-1]) / transformed_data.shape[0]
    else:
        res = cutils.iterate_data_1D_weighted(transformed_data, weights, 
                                              result)
        return np.asfarray(res[:-1])


def linbin_numpy(data, grid_points, weights=None):
    """
    1D Linear binning using NumPy. Assigns weights to grid points from data.

    This function is fast for data sets upto approximately 1-10 million,
    it uses vectorized NumPy functions to perform linear binning. Takes around
    100 ms on 1 million data points, so not nearly as fast as the Cython
    implementation (10 ms).
    
    Parameters
    ----------
    data : array-like
        Should be of shape (obs,).
    grid_points : array-like
        Should be of shape (points,).
    weights : array-like
        Should be of shape (obs,).

    Examples
    --------
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_numpy(data, np.arange(6), weights=None)
    >>> np.allclose(ans, np.array([0, 0, 0.375, 0.375, 0.25, 0]))
    True
    >>> ans = linbin_numpy(data, np.arange(6), weights=np.arange(1, 5))
    >>> np.allclose(ans, np.array([0, 0, 0.2, 0.4, 0.4, 0]))
    True
    >>> data = np.array([2, 2.5, 3, 4])
    >>> ans = linbin_numpy(data, np.arange(1, 7), weights=None)
    >>> np.allclose(ans, np.array([0, 0.375, 0.375, 0.25, 0, 0]))
    True
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=np.float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=np.float)
    assert len(data.shape) == 1

    # Verify that the grid is equidistant
    diffs = np.diff(grid_points)
    assert np.allclose(np.ones_like(diffs) * diffs[0], diffs)

    if weights is None:
        weights = np.ones_like(data)

    weights = np.asarray_chkfinite(weights, dtype=np.float)
    weights = weights / np.sum(weights)

    if not len(data) == len(weights):
        raise ValueError('Length of data must match length of weights.')

    # Transform the data
    min_grid = np.min(grid_points)
    max_grid = np.max(grid_points)
    num_intervals = len(grid_points) - 1  
    dx = (max_grid - min_grid) / num_intervals
    transformed_data = (data - min_grid) / dx

    # Compute the integral and fractional part of the data
    # The integral part is used for lookups, the fractional part is used
    # to weight the data
    fractional, integral = np.modf(transformed_data)
    integral = integral.astype(np.int)

    # Sort the integral values, and the fractional data and weights by
    # the same key. This lets us use binary search, which is faster
    # than using a mask in the the loop below
    indices_sorted = np.argsort(integral)
    integral = integral[indices_sorted]
    fractional = fractional[indices_sorted]
    weights = weights[indices_sorted]

    # Pre-compute these products, as they are used in the loop many times
    frac_weights = fractional * weights
    neg_frac_weights = weights - frac_weights

    # If the data is not a subset of the grid, the integral values will be
    # outside of the grid. To solve the problem, we filter these values away
    unique_integrals = np.unique(integral)
    unique_integrals = unique_integrals[(unique_integrals >= 0) &
                                        (unique_integrals <= len(grid_points))]

    result = np.asfarray(np.zeros(len(grid_points) + 1))
    for grid_point in unique_integrals:

        # Use binary search to find indices for the grid point
        # Then sum the data assigned to that grid point
        low_index = np.searchsorted(integral, grid_point, side='left')
        high_index = np.searchsorted(integral, grid_point, side='right')
        result[grid_point] += neg_frac_weights[low_index:high_index].sum()
        result[grid_point + 1] += frac_weights[low_index:high_index].sum()

    return result[:-1]


def linbin_Ndim_python(data, grid_points, weights=None):
    """
    N-dimensional linear binning. This is a slow, pure-Python function.
    Mainly used for testing purposes.
    
    With :math:`N` data points, and :math:`n` grid points in each dimension
    :math:`d`, the running time is :math:`O(N2^d)`. For each point the
    algorithm finds the nearest points, of which there are two in each
    dimension.
   
    Parameters
    ----------
    data : array-like
        The data must be of shape (obs, dims).
    grid_points : array-like
        Grid, where cartesian product is already performed.
    weights : array-like
        Must have shape (obs,).
        
    Examples
    --------
    >>> from KDEpy.utils import autogrid
    >>> grid_points = autogrid(np.array([[0, 0, 0]]), num_points=(3, 3, 3))
    >>> d = linear_binning(np.array([[1.0, 0, 0]]), grid_points, weights=None)
    """
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=np.float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=np.float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=np.float)
    else:
        # This is not efficient, but this function should just be correct
        weights = np.ones(data.shape[0])
    weights = weights / np.sum(weights)

    if (weights is not None) and (data.shape[0] != len(weights)):
        raise ValueError('Length of data must match length of weights.')
    
    obs_tot, dims = grid_points.shape
    
    # Compute the number of grid points for each dimension in the grid
    grid_num = (grid_points[:, i] for i in range(dims))
    grid_num = np.array(list(len(np.unique(g)) for g in grid_num))
    
    # Scale the data to the grid
    min_grid = np.min(grid_points, axis=0)
    max_grid = np.max(grid_points, axis=0)
    num_intervals = (grid_num - 1)  # Number of intervals
    dx = (max_grid - min_grid) / num_intervals
    data = (data - min_grid) / dx

    # Create results
    result = np.zeros(grid_points.shape[0], dtype=np.float)
        
    # Go through every data point
    for observation, weight in zip(data, weights):
        
        # Compute integer part and fractional part for every x_i
        # Compute relation to previous grid point, and next grid point
        int_frac = (((int(coordinate), 1 - (coordinate % 1)), 
                     (int(coordinate) + 1, (coordinate % 1)))
                    for coordinate in observation)

        # Go through every cartesian product, i.e. every corner in the
        # hypercube grid points surrounding the observation
        for cart_prod in itertools.product(*int_frac):
            
            fractions = (frac for (integral, frac) in cart_prod)
            integrals = list(integral for (integral, frac) in cart_prod) 
            # Find the index in the resulting array, compured by
            # x_1 * (g_2 * g_3 * g_4) + x_2 * (g_3 * g_4) + x_3 * (g_4) + x_4
            
            index = integrals[0]
            for j in range(1, dims):
                index = grid_num[j] * index + integrals[j]
            
            value = functools.reduce(operator.mul, fractions)
            result[index % obs_tot] += value * weight
        
    assert np.allclose(np.sum(result), 1)
    return result


def linbin_Ndim(data, grid_points, weights=None):
    """
    2 and 3-dimensional linear binning.
    
    With :math:`N` data points, and :math:`n` grid points in each dimension
    :math:`d`, the running time is :math:`O(N2^d)`. For each point the
    algorithm finds the nearest points, of which there are two in each
    dimension.
    
    Approximately 200 times faster than pure python implementation.
    
    Parameters
    ----------
    data : array-like
        The data must be of shape (obs, dims).
    grid_points : array-like
        Grid, where cartesian product is already performed.
    weights : array-like
        Must have shape (obs,).
        
    Examples
    --------
    >>> 1 + 1
    2
    """
    data_obs, data_dims = data.shape
    assert len(grid_points.shape) == 2
    assert data_dims >= 2
    
    # Convert the data and grid points
    data = np.asarray_chkfinite(data, dtype=np.float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=np.float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=np.float)
        weights = weights / np.sum(weights)

    if (weights is not None) and (data.shape[0] != len(weights)):
        raise ValueError('Length of data must match length of weights.')
    
    obs_tot, dims = grid_points.shape
    
    # Compute the number of grid points for each dimension in the grid
    grid_num = (grid_points[:, i] for i in range(dims))
    grid_num = np.array(list(len(np.unique(g)) for g in grid_num))
    
    # Scale the data to the grid
    min_grid = np.min(grid_points, axis=0)
    max_grid = np.max(grid_points, axis=0)
    num_intervals = (grid_num - 1)
    dx = (max_grid - min_grid) / num_intervals
    data = (data - min_grid) / dx

    # Create results
    result = np.zeros(grid_points.shape[0], dtype=np.float)
    
    # Call the Cython implementation
    if weights is not None:
        if data_dims >= 3:
            binary_flgs = cartesian(([0, 1], ) * dims)
            result = cutils.iterate_data_ND_weighted(data, weights, result, 
                                                     grid_num, obs_tot, 
                                                     binary_flgs)
        else:
            result = cutils.iterate_data_2D_weighted(data, weights, result, 
                                                     grid_num, obs_tot)
        result = np.asarray_chkfinite(result, dtype=np.float)
    else:
        if data_dims >= 3:
            binary_flgs = cartesian(([0, 1], ) * dims)
            result = cutils.iterate_data_ND(data, result, grid_num, obs_tot, 
                                            binary_flgs)
        else:
            result = cutils.iterate_data_2D(data, result, grid_num, obs_tot)
        result = np.asarray_chkfinite(result, dtype=np.float)
        result = result / data_obs

    assert np.allclose(np.sum(result), 1)
    return result


def linear_binning(data, grid_points, weights=None):
    """
    Compute binning by setting a linear grid and weighting points linearily
    by their distance to the grid points. In addition, weight asssociated with
    data points may be passed.

    Parameters
    ----------
    data
        The data points.
    num_points
        The number of points in the grid.
    weights
        The weights.

    Returns
    -------
    (grid, data)
        Data weighted at each grid point.

    Examples
    --------
    >>> data = [1, 1.5, 1.5, 2, 2.8, 3]
    >>> grid_points = [1, 2, 3]
    >>> data = linear_binning(data, grid_points)
    >>> np.allclose(data, np.array([0.33333, 0.36667, 0.3]))
    True
    """
    data = np.asarray_chkfinite(data, dtype=np.float)
    grid_points = np.asarray_chkfinite(grid_points, dtype=np.float)
    if weights is not None:
        weights = np.asarray_chkfinite(weights, dtype=np.float)

    try:
        obs, dims = data.shape
        
    except ValueError:
        dims = 1
        
    if dims == 1:
        if _use_Cython:
            return linbin_cython(data.ravel(), grid_points, weights=weights)   
        else:
            return linbin_numpy(data.ravel(), grid_points, weights=weights)
    else:
        return linbin_Ndim(data, grid_points, weights=weights)


if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v', '--capture=sys'])
    # pytest.main(args=[__file__, '--doctest-modules', '-v', '--capture=sys'])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for the FFTKDE.
"""
import numbers
import numpy as np
from KDEpy.BaseKDE import BaseKDE
from KDEpy.binning import linear_binning
from scipy.signal import convolve
from KDEpy.utils import cartesian


class FFTKDE(BaseKDE):
    """
    This class implements a convolution (FFT) computation of a KDE. 
    While this implementation is very fast, there are some limitations: (1) the 
    bandwidth must be constant and (2) the KDE must be evaluated on an 
    equidistant grid. The finer the grid, the smaller the error.
    
    The evaluation step is split into two phases. First the :math:`N` data 
    points are binned using a linear binning routine on an equidistant grid `x`
    with :math:`n` grid points. This runs in :math:`O(N 2^d)` time.
    Then the kernel is evaluated once on :math:`\leq n` points and the result 
    of the kernel evaluation and the binned data is convolved. Using the
    convolution theorem, this step runs in :math:`O(n \log n)` time.
    While :math:`N` may be millions, :math:`n` is typically 2**10. The total
    running time of the algorithm is :math:`O(N 2 + n \log n)`. See references.
    
    The implementation is reminiscent of the one found in statsmodels. However,
    ulike the statsmodels implementation every kernel is available for FFT
    computation, weighted data is available for FFT computation, and no large
    temporary arrays are created.

    Parameters
    ----------
    kernel : str
        The kernel function. See cls._available_kernels.keys() for choices.
    bw : float or str
        Bandwidth or bandwidth selection method. If a float is passed, it
        is the standard deviation of the kernel. If a string it passed, it
        is the bandwidth selection method, see cls._bw_methods.keys() for
        choices.
        
    Examples
    --------
    >>> data = np.random.randn(2**10)
    >>> # (1) Automatic bw selection using Improved Sheather Jones
    >>> x, y = FFTKDE(bw='ISJ').fit(data).evaluate()
    >>> # (2) Explicit choice of kernel and bw (standard deviation of kernel)
    >>> x, y = FFTKDE(kernel='triweight', bw=0.5).fit(data).evaluate()
    >>> weights = data + 10
    >>> # (3) Using a grid and weights for the data
    >>> y = FFTKDE(kernel='epa', bw=0.5).fit(data, weights).evaluate(x)
    >>> # (4) If you supply your own grid, it must be equidistant
    >>> y = FFTKDE().fit(data)(np.linspace(-10, 10, num=2**12))
    
    References
    ----------
    - Wand, M. P., and M. C. Jones. Kernel Smoothing. 
      London ; New York: Chapman and Hall/CRC, 1995. Pages 182-192.
    - Statsmodels implementation, at 
      ``statsmodels.nonparametric.kde.KDEUnivariate``.
    """
    
    def __init__(self, kernel='gaussian', bw=1, norm=2):
        self.norm = norm
        super().__init__(kernel, bw)
        assert isinstance(self.norm, numbers.Number) and self.norm > 0
    
    def fit(self, data, weights=None):
        """
        Fit the KDE to the data. This validates the data and stores it. 
        Computations are performed upon evaluation on a grid.
    
        Parameters
        ----------
        data: array-like
            The data points.
        weights: array-like
            One weight per data point. Must have same shape as the data.
            
        Returns
        -------
        self
            Returns the instance.
            
        Examples
        --------
        >>> data = [1, 3, 4, 7]
        >>> weights = [3, 4, 2, 1]
        >>> kde = FFTKDE().fit(data, weights=None)
        >>> kde = FFTKDE().fit(data, weights=weights)
        >>> x, y = kde()
        """
        
        # Sets self.data
        super().fit(data, weights)
        return self
    
    def evaluate(self, grid_points=None):
        """
        Evaluate on equidistant grid points.
        
        Parameters
        ----------
        grid_points: array-like, int, tuple or None
            A grid (mesh) to evaluate on. High dimensional grids must have 
            shape (obs, dims). If an integer is passed, it's the number of grid
            points on an equidistant grid. If a tuple is passed, it's the
            number of grid points in each dimension. If None, a grid will be 
            automatically created.
            
        Returns
        -------
        y: array-like
            If a grid is supplied, `y` is returned. If no grid is supplied,
            a tuple (`x`, `y`) is returned.
            
        Examples
        --------
        >>> kde = FFTKDE().fit([1, 3, 4, 7])
        >>> # Two ways to evaluate, either with a grid or without
        >>> x, y = kde.evaluate()
        >>> x, y = kde.evaluate(256)
        >>> y = kde.evaluate(x)
        """
        
        # This method sets self.grid_points and verifies it
        super().evaluate(grid_points)
        
        if callable(self.bw):
            bw = self.bw(self.data)
        elif isinstance(self.bw, numbers.Number) and self.bw > 0:
            bw = self.bw
        else:
            raise ValueError('The bw must be a callable or a number.')
        self.bw = bw
        
        # Step 1 - Obtaining the grid counts
        data = linear_binning(self.data, grid_points=self.grid_points, 
                              weights=self.weights)
        
        # Step 2 - Computing kernel weights
        g_shape = self.grid_points.shape[1]
        num_grid_points = np.array(list(len(np.unique(self.grid_points[:, i])) 
                                        for i in range(g_shape)))
        
        min_grid = np.min(self.grid_points, axis=0)
        max_grid = np.max(self.grid_points, axis=0)
        num_intervals = (num_grid_points - 1)
        dx = (max_grid - min_grid) / num_intervals
        
        # Find the real bandwidth, the support times the desired bw factor
        if self.kernel.finite_support:
            real_bw = self.kernel.support * self.bw
        else:
            # The parent class should compute this already. If not, compute
            # it again. This optimization only dominates a little bit with
            # few data points
            try:
                real_bw = self._kernel_practical_support
            except AttributeError:
                real_bw = self.kernel.practical_support(self.bw)
            
        # Compute L, the number of dx'es to move out from 0 in kernel
        L = np.minimum(np.floor(real_bw / dx), num_intervals + 1)
        assert (dx * L <= real_bw).all()
        
        # Evaluate the kernel once
        grids = [np.linspace(-dx * L, dx * L, int(L * 2 + 1)) for (dx, L)
                 in zip(dx, L)] 
        kernel_grid = cartesian(grids)
        kernel_weights = self.kernel(kernel_grid, bw=self.bw, norm=self.norm)
        
        # Reshape in preparation to 
        kernel_weights = kernel_weights.reshape(*[int(k * 2 + 1) for k in L])
        data = data.reshape(*tuple(num_grid_points))

        # Step 3 - Performing the convolution
        evaluated = convolve(data, kernel_weights, mode='same').reshape(-1, 1)
        return self._evalate_return_logic(evaluated, self.grid_points)


if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v', '--capture=sys'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  4 20:52:43 2018

@author: tommy
"""

import numpy as np
import collections.abc
import numbers
import functools
from scipy.special import gamma, factorial2
from scipy.stats import norm
from scipy.optimize import brentq

# In R, the following are implemented:
# "gaussian", "rectangular", "triangular", "epanechnikov", 
# "biweight", "cosine" or "optcosine"

# Wikipedia
# uniform, trinagular, epanechnikov, quartic,triweight, tricube,
# gaussian, cosine, logistic, sigmoid, silverman

# All kernel functions take x of shape (obs, dims) and returns (obs, 1)
# All kernel functions integrate to unity


def gauss_integral(n):
    """
    Solve the integral
    \int_0^1 exp(-0.5 * x * x) x^n dx
    
    See
    https://en.wikipedia.org/wiki/List_of_integrals_of_Gaussian_functions
    
    Examples
    --------
    >>> ans = gauss_integral(3)
    >>> np.allclose(ans, 2)
    True
    >>> ans = gauss_integral(4)
    >>> np.allclose(ans, 3.75994)
    True
    """
    factor = np.sqrt(np.pi * 2)
    if n % 2 == 0:
        return factor * factorial2(n - 1) / 2
    elif n % 2 == 1:
        return factor * norm.pdf(0) * factorial2(n - 1)
    else:
        raise ValueError('n must be odd or even.')


def trig_integral(k):
    """
    Returns the solutions to
    Is(k) = int_0^1 sin(pi * x / 2) x^k dx
    Ic(k) = int_0^1 cos(pi * x / 2) x^k dx
    using a recursive formula. Returns a tuple (Is, Ic).
    
    Examples
    --------
    >>> import numpy as np
    >>> Is, Ic = trig_integral(2) # Verify with solution from WolframAlpha
    >>> np.allclose([Is, Ic], [0.29454, 0.12060], rtol=10e-5)
    True
    """

    Ic = 2 / np.pi
    Is = 2 / np.pi
    
    if k <= 0:
        return Is, Ic

    for i in range(1, k + 1):
        Ic, Is = (2 / np.pi) * (1 - i * Is), (2 / np.pi) * i * Ic
        
    return Is, Ic


def p_norm(x, p):
    """
    The p-norm of an array of shape (obs, dims)
    
    Examples
    --------
    >>> x = np.arange(9).reshape((3, 3))
    >>> p = 2
    >>> np.allclose(p_norm(x, p), euclidean_norm(x))
    True
    """
    if np.isinf(p):
        return infinity_norm(x)
    elif p == 2:
        return euclidean_norm(x)
    elif p == 1:
        return taxicab_norm(x)
    return np.power(np.power(np.abs(x), p).sum(axis=1), 1 / p)


def euclidean_norm(x):
    """
    The 2 norm of an array of shape (obs, dims)
    """
    return np.sqrt((x * x).sum(axis=1))


def euclidean_norm_sq(x):
    """
    The squared 2 norm of an array of shape (obs, dims)
    """
    return (x * x).sum(axis=1)


def infinity_norm(x):
    """
    The infinity norm of an array of shape (obs, dims)
    """
    return np.abs(x).max(axis=1)


def taxicab_norm(x):
    """
    The taxicab norm of an array of shape (obs, dims)
    """
    return np.abs(x).sum(axis=1)


def volume_unit_ball(d, p=2):
    """
    Volume of d-dimensional unit ball under the p-norm. When p=1 this is called
    a cross-polytype, when p=2 it's called a hypersphere, and when p=infty it's
    called a hypercube. 
    
    Notes
    -----
    See the following paper for a very general result related to this:
        
    - Wang, Xianfu. “Volumes of Generalized Unit Balls.” 
      Mathematics Magazine 78, no. 5 (2005): 390–95. 
      https://doi.org/10.2307/30044198.
    """
    return 2.**d * gamma(1 + 1 / p) ** d / gamma(1 + d / p)


def epanechnikov(x, dims=1):
    normalization = (2 / (dims + 2))
    dist_sq = x**2 
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = (1 - dist_sq)[mask] / normalization
    return out


def gaussian(x, dims=1):
    normalization = dims * gauss_integral(dims - 1)
    dist_sq = x**2
    return np.exp(-dist_sq / 2) / normalization


def box(x, dims=1):
    normalization = 1
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = 1 / normalization
    return out


def exponential(x, dims=1):
    normalization = gamma(dims) * dims
    return np.exp(-x) / normalization


def tri(x, dims=1):
    normalization = (1 / (dims + 1))
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.maximum(0, 1 - x)[mask] / normalization 
    return out


def biweight(x, dims=1):
    normalization = (8 / ((dims + 2) * (dims + 4)))
    dist_sq = x**2
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = np.maximum(0, (1 - dist_sq)**2)[mask] / normalization
    return out


def triweight(x, dims=1):
    normalization = (48 / ((dims + 2) * (dims + 4) * (dims + 6)))
    dist_sq = x**2
    out = np.zeros_like(dist_sq)
    mask = dist_sq < 1
    out[mask] = np.maximum(0, (1 - dist_sq)**3)[mask] / normalization
    return out


def tricube(x, dims=1):
    normalization = (162 / ((dims + 3) * (dims + 6) * (dims + 9)))
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.maximum(0, (1 - x**3)**3)[mask] / normalization
    return out


def cosine(x, dims=1):
    Is, Ic = trig_integral(dims - 1)
    normalization = Ic
    out = np.zeros_like(x)
    mask = x < 1
    out[mask] = np.cos((np.pi * x) / 2)[mask] / (normalization * dims)
    return out


def logistic(x, dims=1):
    return 1 / (2 + 2 * np.cosh(x))


def sigmoid(x, dims=1):
    return (1 / (np.pi * np.cosh(x)))


class Kernel(collections.abc.Callable):
    
    def __init__(self, function, var=1, support=3):
        """
        Initialize a new kernel function.
        
        function: callable, numpy.arr -> numpy.arr, should integrate to 1
        expected_value : peak, typically 0
        support: support of the function.
        """
        self.function = function
        self.var = var
        self.finite_support = np.isfinite(support)
        
        # If the function has finite support, scale the support so that it
        # corresponds to the support of the function when it is scaled to have
        # unit variance.
        self.support = support / np.sqrt(self.var)
        
    def practical_support(self, bw, atol=10e-5):
        """
        Return the support for practical purposes.
        """
        # If the kernel has finite support, return the support accounting for
        # the bw
        if self.finite_support:
            return self.support * bw
        
        # If the function does not have finite support, find a practical value
        else:
            def f(x):
                return self.evaluate(x, bw=bw) - atol
            try:
                xtol = 1e-3
                ans = brentq(f, a=0, b=8 * bw, xtol=xtol, full_output=False)
                return ans + xtol
            except ValueError:
                msg = 'Unable to solve for support numerically. Use a ' + \
                      'kernel with finite support or scale data to smaller bw.'
                raise ValueError(msg)
    
    def evaluate(self, x, bw=1, norm=2):
        """
        Evaluate the kernel.
        
        Parameters
        ----------
        x : array-like
            Must have shape (obs, dims).
        bw : array-like
            Must have shape (obs, ), or float.
        """
        
        # If x is a number, convert it to a length-1 NumPy vector
        if isinstance(x, numbers.Number):
            x = np.asarray_chkfinite([x])
        else:
            x = np.asarray_chkfinite(x)
            
        if len(x.shape) == 1:
            x = x.reshape(-1, 1)
            
        # Scale the function, such that bw=1 corresponds to the function having
        # a standard deviation (or variance) equal to 1
        real_bw = bw / np.sqrt(self.var)
        obs, dims = x.shape
        
        # Set the volume function
        volume_func = functools.partial(volume_unit_ball, p=norm)
            
        if dims > 1:
            distances = p_norm(x, norm).ravel()
        else:
            distances = np.abs(x).ravel()
            
        return (self.function(distances / real_bw, dims) / 
                ((real_bw**dims) * volume_func(dims)))
            
    __call__ = evaluate
    
    
gaussian = Kernel(gaussian, var=1, support=np.inf)
exp = Kernel(exponential, var=4, support=np.inf)
box = Kernel(box, var=1 / 3, support=1)
tri = Kernel(tri, var=1 / 6, support=1)
epa = Kernel(epanechnikov, var=1 / 5, support=1)
biweight = Kernel(biweight, var=1 / 7, support=1)
triweight = Kernel(triweight, var=1 / 9, support=1)
tricube = Kernel(tricube, var=35 / 243, support=1)
cosine = Kernel(cosine, var=(1 - (8 / np.pi**2)), support=1)
logistic = Kernel(logistic, var=(np.pi**2 / 3), support=np.inf)
sigmoid = Kernel(sigmoid, var=(np.pi**2 / 4), support=np.inf)

_kernel_functions = {'gaussian': gaussian,
                     'exponential': exp,
                     'box': box,
                     'tri': tri,
                     'epa': epa,
                     'biweight': biweight,
                     'triweight': triweight,
                     'tricube': tricube,
                     'cosine': cosine,
                     # 'logistic': logistic,
                     # 'sigmoid': sigmoid
                     }


if __name__ == "__main__" and False:
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v'])
    
    plot = False
    if plot:
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D
        a = Axes3D
        import scipy
        for name, func in _kernel_functions.items():
    
            print('-' * 2**7)
            print(name)
            print(func([-1, 0, 1]))
            print(func(np.array([[0, -1], [0, 0], [0, 1]])))
            print(func(np.array([[0, -1, 0], [0, 0, 0], [0, 1, 0]])))
            
            # Plot in 1D
            n = 50
            x = np.linspace(-3, 3, num=n * 3)
            plt.plot(x, func(x))
            plt.show()
            
            # Plot in 2D
            n = 50
            linspace = np.linspace(-3, 3, num=n)
    
            x, y = linspace, linspace
            k = np.array(np.meshgrid(x, y)).T.reshape(-1, 2)
            z = func(k).reshape((n, n))
            
            x, y = np.meshgrid(x, y)
            
            fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))
            
            surf = ax.plot_surface(x, y, z, rstride=1, cstride=1,
                                   linewidth=1, antialiased=True, shade=True)
            
            angle = 90
            ax.view_init(30, angle)
    
            plt.show()
            
            # Perform integration 1D
            def int1D(x1):
                return func(x1)
            
            ans, err = scipy.integrate.nquad(int1D, [[-4, 4]])
            print('1D integration result: {}'.format(ans))
            assert np.allclose(ans, 1, rtol=10e-3, atol=10e-3)
            
            # Perform integration 2D
            def int2D(x1, x2):
                return func([[x1, x2]])
            
            ans, err = scipy.integrate.nquad(int2D, [[-4, 4], [-4, 4]],
                                             opts={'epsabs': 10e-3, 
                                                   'epsrel': 10e-3})
            print('2D integration result: {}'.format(ans))
            assert np.allclose(ans, 1, rtol=10e-3, atol=10e-3)
        
if __name__ == "__main__" and False:
    
    bw = 2
    print(gaussian.practical_support(bw))
    print(epa.practical_support(bw))
    import matplotlib.pyplot as plt
    x = np.linspace(-10, 10, num=2**8)
    y = gaussian(x, bw=bw)
    plt.plot(x, y)
    plt.scatter([-gaussian.practical_support(bw), 
                 gaussian.practical_support(bw)], [0, 0])
    
    y = epa(x, bw=bw)
    plt.plot(x, y)
    plt.scatter([-epa.practical_support(bw), 
                 epa.practical_support(bw)], [0, 0])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  4 10:52:17 2018

@author: tommy
"""
from scipy.spatial import cKDTree
import numbers
import numpy as np
import time
from KDEpy.BaseKDE import BaseKDE

   
class TreeKDE(BaseKDE):
    """
    This class implements a tree-based computation of a kernel density 
    estimate. It works by segmenting the space recursively into smaller parts. 
    This makes computing a kernel density estimate at a location easier, since 
    we are able to query the tree structure for nearby points instead of having 
    to evaluate the kernel function on all data points. For kernels without 
    finite support, their support is approximated. The ``scipy`` k-d tree is
    used as the underlying algorithm.
    
    Parameters
    ----------
    kernel : str
        The kernel function. See cls._available_kernels.keys() for choices.
    bw : float, str or array-like
        Bandwidth or bandwidth selection method. If a float is passed, it
        is the standard deviation of the kernel. If a string it passed, it
        is the bandwidth selection method, see cls._bw_methods.keys() for
        choices. If an array-like it passed, it is the bandwidth of each
        point.
    norm : float
        The p-norm used to compute the distances in higher dimensions.
        
    Examples
    --------
    >>> data = np.random.randn(2**10)
    >>> # (1) Automatic bw selection using Improved Sheather Jones
    >>> x, y = TreeKDE(bw='ISJ').fit(data).evaluate()
    >>> # (2) Explicit choice of kernel and bw (standard deviation of kernel)
    >>> x, y = TreeKDE(kernel='triweight', bw=0.5).fit(data).evaluate()
    >>> weights = data + 10
    >>> # (3) Using a grid and weights for the data
    >>> y = TreeKDE(kernel='epa', bw=0.5).fit(data, weights).evaluate(x)
    
    References
    ----------
    - Friedman, Jerome H., Jon Louis Bentley, and Raphael Ari Finkel. 
      An Algorithm for Finding Best Matches in Logarithmic Expected Time.
      ACM Trans. Math. Softw. 3, no. 3 (September 1977): 209–226. 
      https://doi.org/10.1145/355744.355745.
    - Maneewongvatana, Songrit, and David M. Mount. 
      It’s Okay to Be Skinny, If Your Friends Are Fat.
      In Center for Geometric Computing 4th Annual Workshop on Computational 
      Geometry, 2:1–8, 1999.
    - Silverman, B. W. Density Estimation for Statistics and Data Analysis. 
      Boca Raton: Chapman and Hall, 1986. Page 99 for reference to kd-tree.
    - Scipy implementation, at ``scipy.spatial.KDTree``.
    """
    
    def __init__(self, kernel='gaussian', bw=1, norm=2.):
        super().__init__(kernel, bw)
        self.norm = norm
    
    def fit(self, data, weights=None):
        """
        Fit the KDE to the data. This validates the data and stores it. 
        Computations are performed upon evaluation on a grid.
    
        Parameters
        ----------
        data: array-like
            The data points.
        weights: array-like
            One weight per data point. Must have same shape as the data.
            
        Returns
        -------
        self
            Returns the instance.
            
        Examples
        --------
        >>> data = [1, 3, 4, 7]
        >>> weights = [3, 4, 2, 1]
        >>> kde = TreeKDE().fit(data, weights=None)
        >>> kde = TreeKDE().fit(data, weights=weights)
        >>> x, y = kde()
        """
        # Sets self.data
        super().fit(data, weights)
        return self
    
    def evaluate(self, grid_points=None, eps=10e-4):
        """
        Evaluate on grid points.
        
        Parameters
        ----------
        grid_points: array-like, int, tuple or None
            A grid (mesh) to evaluate on. High dimensional grids must have 
            shape (obs, dims). If an integer is passed, it's the number of grid
            points on an equidistant grid. If a tuple is passed, it's the
            number of grid points in each dimension. If None, a grid will be 
            automatically created.
        eps: float
            The maximal total error in absolute terms when estimating the 
            effective support of a kernel which has infinite support. Setting
            this too high will produced a jagged estimate.
            
        Returns
        -------
        y: array-like
            If a grid is supplied, `y` is returned. If no grid is supplied,
            a tuple (`x`, `y`) is returned.
            
        Examples
        --------
        >>> kde = TreeKDE().fit([1, 3, 4, 7])
        >>> # Two ways to evaluate, either with a grid or without
        >>> x, y = kde.evaluate()
        >>> x, y = kde.evaluate(256)
        >>> y = kde.evaluate(x)
        """
        
        # This method sets self.grid points and verifies it
        super().evaluate(grid_points)
        
        evaluated = np.zeros(self.grid_points.shape[0])
        
        # For every data point, compute the kernel and add to the grid
        obs, dims = self.data.shape
        bw = self.bw
        if isinstance(bw, numbers.Number):
            bw = np.asfarray(np.ones(obs) * bw)
        elif callable(bw):
            bw = np.asfarray(np.ones(obs) * bw(self.data))
        else:
            bw = np.asarray_chkfinite(bw, dtype=np.float)

        # Initialize the tree structure for fast lookups of neighbors
        tree = cKDTree(self.data)
        
        # Compute the kernel radius
        maximal_bw = np.max(bw)
        if not eps > 0:
            raise ValueError('eps must be > 0.')
        kernel_radius = self.kernel.practical_support(maximal_bw, eps)
            
        # Since we iterate through grid points, we need the maximum bw to
        # ensure that we get data points that are close enough
        for i, grid_point in enumerate(self.grid_points):

            # Query for data points that are close to this grid point
            # TODO: Is this epsilon value sensible?
            indices = tree.query_ball_point(x=grid_point, r=kernel_radius,
                                            p=self.norm, eps=eps * obs**0.5)

            # Use broadcasting to find x-values (distances)
            x = grid_point - self.data[indices]
            kernel_estimates = self.kernel(x, bw=bw[indices], norm=self.norm)
            
            if self.weights is not None:
                weights_subset = self.weights[indices]
                assert kernel_estimates.shape == weights_subset.shape
            assert kernel_estimates.shape == bw[indices].shape

            # Unpack the (n, 1) arrays to (n,) and compute the doc product
            if self.weights is not None:
                evaluated[i] += np.dot(kernel_estimates, weights_subset)
            else:
                evaluated[i] += np.sum(kernel_estimates) / obs

        return self._evalate_return_logic(evaluated, self.grid_points)


if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v'])

if __name__ == '__main__':
    
    import matplotlib.pyplot as plt
    from KDEpy.NaiveKDE import NaiveKDE
    
    # Comparing tree and naive
    # -----------------------------------------
    data = [3, 3.5, 4, 6, 8]
    kernel = 'triweight'
    bw = [3, 0.3, 1, 0.3, 2]
    weights = [1, 1, 1, 1, 1]
    
    plt.figure(figsize=(10, 4))
    plt.title('Basic example of the naive KDE')
    
    plt.subplot(1, 2, 1)
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(0, 10, num=1024)
    for d, b in zip(data, bw):
        k = NaiveKDE(kernel=kernel, bw=b).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    
    plt.subplot(1, 2, 2)
    kde = TreeKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(0, 10, num=1024)
    for d, b in zip(data, bw):
        k = NaiveKDE(kernel=kernel, bw=b).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    y = kde.evaluate(x)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    plt.show()

    # Comparing tree and naive
    # -----------------------------------------
    data = [3, 3.5, 4, 6, 8]
    data = np.array(data)
    data = list(np.random.random(5))
    kernel = 'triweight'
    bw = [3, 0.3, 1, 0.3, 2]
    weights = [1, 2, 3, 4, 5]
    np.random.seed(123)
    n = 120
    data = np.random.gamma(5, scale=2.5, size=n) * 10
    bw = np.random.random(n) + 5 + 1
    weights = np.random.random(n) / 1 + 1
    
    plt.figure(figsize=(10, 4))
    plt.title('Basic example of the naive KDE')
    
    plt.subplot(1, 2, 1)
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(-4, 360, num=1024)
    for d, b in zip(data, bw):
        break
        k = NaiveKDE(kernel=kernel, bw=b).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    st = time.perf_counter()
    y = kde.evaluate(x)
    print(time.perf_counter() - st)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    
    plt.subplot(1, 2, 2)
    kde = TreeKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(-4, 360, num=1024)
    for d, b in zip(data, bw):
        break
        k = NaiveKDE(kernel=kernel, bw=b).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    st = time.perf_counter()
    y = kde.evaluate(x)
    print(time.perf_counter() - st)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))

    plt.show()
    
    # Comparing tree and naive
    # -----------------------------------------
    kernel = 'gaussian'
    data = np.sqrt(np.arange(5) + 1)
    # data = np.ones(100)
    bw = np.ones_like(data)
    weights = None
    eps = 0.0001

    plt.figure(figsize=(10, 4))
    plt.title('Basic example of the naive KDE')
    
    plt.subplot(1, 2, 1)
    kde = NaiveKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(-4, 4, num=1024)
    for d, b in zip(data, bw):
        break
        k = NaiveKDE(kernel=kernel, bw=b).fit([d]).evaluate(x) / len(data)
        plt.plot(x, k, color='k', ls='--')
        
    st = time.perf_counter()
    x, y = kde.evaluate()
    print(time.perf_counter() - st)
    plt.plot(x, y)
    plt.scatter(data, np.zeros_like(data))
    
    plt.subplot(1, 2, 2)
    kde = TreeKDE(kernel=kernel, bw=bw)
    kde.fit(data, weights)
    x = np.linspace(-4, 4, num=1024)
    for d, b in zip(data, bw):
        break
        k = (NaiveKDE(kernel=kernel, bw=b).fit([d])
             .evaluate(x, eps=eps) / len(data))
        plt.plot(x, k, color='k', ls='--')
        
    st = time.perf_counter()
    x, y = kde.evaluate(eps=eps)
    print(time.perf_counter() - st)
    plt.plot(x, y)
    plt.plot([min(data) - 1, max(data) + 1], [eps, eps])
    plt.scatter(data, np.zeros_like(data))

    plt.show()
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages, Extension
# To use a consistent encoding
from codecs import open
from os import path
# from KDEpy import __version__
VERSION = '0.6' # __version__


def read(fname):
    return open(path.join('.', fname)).read()


try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True


try:
    import numpy as np
except ImportError:
    use_numpy = False
else:
    use_numpy = True

cmdclass = {}
ext_modules = []

if use_cython and use_numpy:
    ext_modules += [Extension("cutils", [path.join('KDEpy', 'cutils.pyx')]),]
    cmdclass.update({ 'build_ext': build_ext })
    include_dirs = [np.get_include()]
else:
    ext_modules += [Extension("cutils", [path.join('KDEpy', 'cutils.c')]),]
    include_dirs = []


setup(
    name='KDEpy',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,

    description='Kernel Density Estimation in Python.',
    long_description='Kernel Density Estimation in Python.',
    # read('README.rst'),
    # The project's main homepage.
    url='https://github.com/tommyod/KDEpy',

    # Author details
    author='tommyod',
    author_email='tod001@uib.no',
    # Choose your license
    license='GNU GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        # 'Intended Audience :: End Users/Desktop',
        # 'Intended Audience :: Healthcare Industry',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages('.', exclude=['contrib', 'docs', 'tests']),
    package_dir={'KDEpy': 'KDEpy'},

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
    'numpy>=1.14.2',
    'scipy>=1.0.1',
    'setuptools>=39.2.0'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # include_package_data=True,
    package_data={
        '': ['templates/*', '*.tex', '*.html'],
    },

    # For cython, see: http://cython.readthedocs.io/en/latest/src/tutorial/cython_tutorial.html
    # ext_modules = cythonize(path.join(".", "KDEpy", "cutils.pyx")),
    cmdclass=cmdclass,
    include_dirs=include_dirs,
    ext_modules=ext_modules,

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #    'console_scripts': [
    #         'sample=sample:main',
    #    ],
    # }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for the BaseKDE class.
"""
from abc import ABC, abstractmethod
from collections.abc import Sequence
import numbers
import numpy as np
from KDEpy.kernel_funcs import _kernel_functions
from KDEpy.bw_selection import _bw_methods
from KDEpy.utils import autogrid


class BaseKDE(ABC):
    """
    Abstract Base Class for every kernel density estimator.
    
    This class is never instantiated, it merely defines some common methods
    which every subclass must implement. In summary, it facilitates:
        
        - The `_available_kernels` parameter
        - Correct handling of `kernel` and `bw` in __init__
        - Forces subclasses to implement `fit(data)`, converts `data` to 
          correct shape (obs, dims) and converts `weights` to correct shape
          (obs,)
        - Forces subclasses to implement `evaluate(grid_points)`, with handling
    """
    
    _available_kernels = _kernel_functions
    _bw_methods = _bw_methods
    
    @abstractmethod
    def __init__(self, kernel: str, bw: float):
        """Initialize the kernel density estimator.

        The return type must be duplicated in the docstring to comply
        with the NumPy docstring style.
    
        Parameters
        ----------
        kernel
            Kernel function, or string matching available options.
        bw
            The bandwidth, either a number, a string or an array-like.
        """
        
        # Verify that the choice of kernel is valid, and set the function
        akernels = sorted(list(self._available_kernels.keys()))
        msg = 'Kernel must be a string or callable. Opts: {}'.format(akernels)
        if isinstance(kernel, str):
            kernel = kernel.strip().lower()
            if kernel not in akernels:
                raise ValueError(msg)
            else:
                self.kernel = self._available_kernels[kernel]
        elif callable(kernel):
            self.kernel = kernel
        else:
            raise ValueError(msg)
        
        # The `bw` paramter may either be a positive number, a string, or 
        # array-like such that each point in the data has a uniue bw
        if (isinstance(bw, numbers.Number) and bw > 0):
            self.bw = bw
        elif isinstance(bw, str):
            amethods = sorted(list(self._bw_methods.keys()))
            if bw.lower() not in set(m.lower() for m in amethods):
                msg = 'bw not recognized. Options are: {}'.format(amethods)
                raise ValueError(msg)
            self.bw = self._bw_methods[bw]
        elif isinstance(bw, (np.ndarray, Sequence)):
            self.bw = bw
        else:
            raise ValueError('Bandwidth must be > 0, array-like or a string.')
            
        # Test quickly that the method has done what is was supposed to do
        assert callable(self.kernel)
        assert (isinstance(self.bw, (np.ndarray, Sequence, numbers.Number)) or 
                callable(self.bw))
            
    @abstractmethod
    def fit(self, data, weights=None):
        """
        Fit the kernel density estimator to the data.
        This method converts the data to shape (obs, dims) and the weights
        to (obs,).

        Parameters
        ----------
        data : array-like or Sequence
            May be array-like of shape (obs,), shape (obs, dims) or a
            Python Sequence, e.g. a list or tuple.
        weights : array-like, Sequence or None
            May be array-like of shape (obs,), shape (obs, dims), a 
            Python Sequence, e.g. a list or tuple, or None.
        """
        
        # -------------- Set up the data depending on input -------------------
        # In the end, the data should be an ndarray of shape (obs, dims)
        data = self._process_sequence(data)
            
        obs, dims = data.shape

        if not obs > 0:
            raise ValueError('Data must contain at least one data point.')
        assert dims > 0
        self.data = data
        
        # -------------- Set up the weights depending on input ----------------
        if weights is not None:
            self.weights = self._process_sequence(weights).ravel()
            self.weights = self.weights / np.sum(self.weights)
            if not obs == len(self.weights):
                raise ValueError('Number of data obs must match weights')
        else:
            self.weights = weights
                
        # Test quickly that the method has done what is was supposed to do
        assert len(self.data.shape) == 2
        if self.weights is not None:
            assert len(self.weights.shape) == 1
            assert self.data.shape[0] == len(self.weights)
    
    @abstractmethod
    def evaluate(self, grid_points=None, bw_to_scalar=True):
        """
        Evaluate the kernel density estimator on the grid points.
        
        Parameters
        ----------
        grid_points : integer, tuple or array-like
            If an integer, the number of equidistant grid point in every
            dimension. If a tuple, the number of grid points in each
            dimension. If array-like, grid points of shape (obs, dims).
        """
        if not hasattr(self, 'data'):
            raise ValueError('Must call fit before evaluating.')
            
        # -------------- Set up the bandwidth depending on inputs -------------
        if isinstance(self.bw, (np.ndarray, Sequence)):
            if bw_to_scalar:
                bw = max(self.bw)
            else:
                bw = self.bw
        elif callable(self.bw):
            bw = self.bw(self.data)
        else:
            bw = self.bw
        self.bw = bw
            
        # -------------- Set up the grid depending on input -------------------
        # If the grid None or an integer, use that in the autogrid method
        types_for_autogrid = (numbers.Integral, tuple)
        if grid_points is None or isinstance(grid_points, types_for_autogrid):
            self._user_supplied_grid = False
            bw_grid = self.kernel.practical_support(bw)
            grid_points = autogrid(self.data, bw_grid, grid_points)
            # Set it here, so as not to call kernel.practical_support(bw) again
            self._kernel_practical_support = bw_grid
        else:
            self._user_supplied_grid = True
            grid_points = self._process_sequence(grid_points)
                
        obs, dims = grid_points.shape
        if not obs > 0:
            raise ValueError('Grid must contain at least one data point.') 
        self.grid_points = grid_points
    
        # Test quickly that the method has done what is was supposed to do
        if bw_to_scalar:
            assert isinstance(self.bw, numbers.Number)
            assert self.bw > 0
        assert len(self.grid_points.shape) == 2
           
    @staticmethod
    def _process_sequence(sequence_array_like):
        """
        Process a sequence of data input to ndarray of shape (obs, dims).

        Parameters
        ----------
        sequence_array_like : Sequence or array-like
            The input data.

        Examples
        --------
        >>> res = BaseKDE._process_sequence([1, 2, 3]) 
        >>> (res == np.array([[1], [2], [3]])).all()
        True
        """
        # Must convert to float to avoid possible interger overflow
        if isinstance(sequence_array_like, Sequence):
            out = np.asfarray(sequence_array_like).reshape(-1, 1)
        elif isinstance(sequence_array_like, np.ndarray):
            if len(sequence_array_like.shape) == 1:
                out = sequence_array_like.reshape(-1, 1)
            elif len(sequence_array_like.shape) == 2:
                out = sequence_array_like
            else:
                raise ValueError('Must be of shape (obs, dims)')
        else:
            raise TypeError('Must be of shape (obs, dims)')
        return np.asarray_chkfinite(out, dtype=np.float)
        
    def _evalate_return_logic(self, evaluated, grid_points):
        """
        Return either evaluation points y, or tuple (x, y) based on inputs.
        """
        # Adding epsilon to output helps contour plotting functions
        evaluated = evaluated.ravel() + np.finfo(float).eps
        obs, dims = grid_points.shape
        if self._user_supplied_grid:
            return evaluated
        else:
            if dims == 1:
                return grid_points.ravel(), evaluated
            return grid_points, evaluated
  
    def __call__(self, *args, **kwargs):
        return self.evaluate(*args, **kwargs)


if __name__ == "__main__":
    import pytest
    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=['.', '--doctest-modules', '-v'])
